import cga_py as ga
from cga_py import dorst_motions as dm
import numpy as np


def test_iso_scale():
    a = np.array(np.random.rand(3))
    np.testing.assert_array_almost_equal(
        dm.iso_scale(a).coeff, (ga.point(a) ^ ga.e_i).coeff
    )
    b = ga.rand_point()
    np.testing.assert_array_almost_equal(dm.iso_scale(b).coeff, (b ^ ga.e_i).coeff)


def test_transv():
    a = np.array([1, 2, 3])
    b = np.array([1, 2, 3])
    d = np.sqrt(14)
    with np.testing.assert_raises(ga.GeometryError):
        dm.transv(ga.point([1, 2, 3]), ga.plane([1, 2, 3], 2))
    x = dm.transv(a, b).coeff
    x /= x[0]
    y = (ga.point(a) ^ ga.plane(b, d)).coeff
    y /= y[0]
    np.testing.assert_array_almost_equal(x, y)
    x = dm.transv(ga.point(a), b).coeff
    x /= x[0]
    y = (ga.point(a) ^ ga.plane(b, d)).coeff
    y /= y[0]
    np.testing.assert_array_almost_equal(x, y)
    x = dm.transv(ga.point(a), ga.plane(b, d)).coeff
    x /= x[0]
    y = (ga.point(a) ^ ga.plane(b, d)).coeff
    y /= y[0]
    np.testing.assert_array_almost_equal(x, y)
    x = dm.transv(a, ga.plane(b, d)).coeff
    x /= x[0]
    y = (ga.point(a) ^ ga.plane(b, d)).coeff
    y /= y[0]
    np.testing.assert_array_almost_equal(x, y)



import cga_py as ga
import numpy as np


def test_act():
    r = ga.cga_object(np.random.rand(32))
    a = ga.cga_object(np.random.rand(32))
    np.testing.assert_array_almost_equal(ga.act(r, a).coeff, (r * a * (~r)).coeff)


def test_com():
    r = ga.cga_object(np.random.rand(32))
    a = ga.cga_object(np.random.rand(32))
    np.testing.assert_array_almost_equal(
        ga.com(r, a).coeff, (0.5 * (r * a - a * r)).coeff
    )


def test_anti_com():
    r = ga.cga_object(np.random.rand(32))
    a = ga.cga_object(np.random.rand(32))
    np.testing.assert_array_almost_equal(
        ga.anti_com(r, a).coeff, (0.5 * (r * a + a * r)).coeff
    )


def test_r_norm():
    r = ga.cga_object(np.random.rand(32))
    np.testing.assert_array_almost_equal(ga.r_norm(r).coeff, (r * (~r)).coeff)


def test_l_norm():
    r = ga.cga_object(np.random.rand(32))
    np.testing.assert_array_almost_equal(ga.l_norm(r).coeff, ((~r) * r).coeff)


def test_grade():
    a0 = np.zeros(32, dtype=complex)
    a1 = np.zeros(32, dtype=complex)
    a2 = np.zeros(32, dtype=complex)
    a3 = np.zeros(32, dtype=complex)
    a4 = np.zeros(32, dtype=complex)
    a5 = np.zeros(32, dtype=complex)
    a = ga.cga_object(np.random.rand(32))
    grade_indices = np.array(
        [
            [0],
            np.arange(1, 6),
            np.arange(6, 16),
            np.arange(16, 26),
            np.arange(26, 31),
            31,
        ],
        dtype=object,
    )

    a0[grade_indices[0]] = a.coeff[grade_indices[0]]
    a1[grade_indices[1]] = a.coeff[grade_indices[1]]
    a2[grade_indices[2]] = a.coeff[grade_indices[2]]
    a3[grade_indices[3]] = a.coeff[grade_indices[3]]
    a4[grade_indices[4]] = a.coeff[grade_indices[4]]
    a5[grade_indices[5]] = a.coeff[grade_indices[5]]
    b0 = ga.n_grade(a, 0).coeff
    b1 = ga.n_grade(a, 1).coeff
    b2 = ga.n_grade(a, 2).coeff
    b3 = ga.n_grade(a, 3).coeff
    b4 = ga.n_grade(a, 4).coeff
    b5 = ga.n_grade(a, 5).coeff
    np.testing.assert_array_almost_equal(a0, b0)
    np.testing.assert_array_almost_equal(a1, b1)
    np.testing.assert_array_almost_equal(a2, b2)
    np.testing.assert_array_almost_equal(a3, b3)
    np.testing.assert_array_almost_equal(a4, b4)
    np.testing.assert_array_almost_equal(a5, b5)



import cga_py as ga
import numpy as np


def test_point_array_error():
    with np.testing.assert_raises(ValueError):
        ga.point(np.random.rand(2))
    with np.testing.assert_raises(ValueError):
        ga.point(np.random.rand(4))
    with np.testing.assert_raises(ValueError):
        ga.point(np.random.rand(1))
    with np.testing.assert_raises(ValueError):
        ga.point(np.random.rand(0))


def test_normalize_point_error():
    a = ga.rand_point()
    b = a + np.random.rand()
    c = a + ga.cga_object(np.append(np.zeros(6), np.random.rand(32 - 6)))
    with np.testing.assert_raises(ValueError):
        ga.normalize_point(b)
    with np.testing.assert_raises(ValueError):
        ga.normalize_point(c)


def test_sphere_array_error():
    with np.testing.assert_raises(ValueError):
        ga.sphere(np.random.rand(2), 5)
    with np.testing.assert_raises(ValueError):
        ga.sphere(np.random.rand(4), 6)
    with np.testing.assert_raises(ValueError):
        ga.sphere(np.random.rand(1), 7)
    with np.testing.assert_raises(ValueError):
        ga.sphere(np.random.rand(0), 8)
    with np.testing.assert_raises(TypeError):
        ga.sphere(np.random.rand(3))
    with np.testing.assert_raises(TypeError):
        ga.sphere(np.random.rand(2))
    with np.testing.assert_raises(TypeError):
        ga.sphere(np.random.rand(4))
    with np.testing.assert_raises(TypeError):
        ga.sphere(np.random.rand(1))
    with np.testing.assert_raises(TypeError):
        ga.sphere(np.random.rand(0))


def test_normalize_sphere_error():
    a = ga.rand_sphere()
    b = a + np.random.rand()
    c = a + ga.cga_object(np.append(np.zeros(6), np.random.rand(32 - 6)))
    with np.testing.assert_raises(ValueError):
        ga.normalize_sphere(b)
    with np.testing.assert_raises(ValueError):
        ga.normalize_sphere(c)


def test_plane_array_error():
    with np.testing.assert_raises(ValueError):
        ga.plane(np.random.rand(2), 5)
    with np.testing.assert_raises(ValueError):
        ga.plane(np.random.rand(4), 6)
    with np.testing.assert_raises(ValueError):
        ga.plane(np.random.rand(1), 7)
    with np.testing.assert_raises(ValueError):
        ga.plane(np.random.rand(0), 8)
    with np.testing.assert_raises(TypeError):
        ga.plane(np.random.rand(3))
    with np.testing.assert_raises(TypeError):
        ga.plane(np.random.rand(2))
    with np.testing.assert_raises(TypeError):
        ga.plane(np.random.rand(4))
    with np.testing.assert_raises(TypeError):
        ga.plane(np.random.rand(1))
    with np.testing.assert_raises(TypeError):
        ga.plane(np.random.rand(0))


def test_normalize_plane_error():
    a = ga.rand_plane()
    b = a + np.random.rand()
    c = a + ga.cga_object(np.append(np.zeros(5), np.random.rand(32 - 5)))
    with np.testing.assert_raises(ValueError):
        ga.normalize_plane(b)
    with np.testing.assert_raises(ValueError):
        ga.normalize_plane(c)



from setuptools import setup

me = "Daren Thimm"
my_email = "daren.thimm@gmx.at"
setup(
    name="cga-py",
    version="1.1.1",
    packages=[
        "cga_py",
    ],
    scripts=["cga_py/bin/cga_py_vis.py"],
    license="BSD-3-Clause License",
    url="https://github.com/physicspenguin/cga-py",
    author=me,
    author_email=my_email,
    maintainer=me,
    maintainer_email=my_email,
    description="Implementation of CGA for python",
    long_description=open("./README.md").read(),
    long_description_content_type="text/markdown",
)



from .kinematic_polynomials import poly_act
from .geom_generators import point, point_to_cartesian
import numpy as np
import multiprocessing as mp
from functools import partial
from numba import njit
import pyqtgraph.opengl as gl


def point_p_act_helper(i, param, poly, points):
    return point_to_cartesian(poly_act(param, poly, point(points[i])))


def point_p_act(param, poly, points):
    """Use poly_act on an array of points given in cartesian coordinates.

    Parameters
    ----------
    param: float
        parameter of polynomial
    poly: ndarray
        polynomial given as coefficient list with increasing degree
    points : ndarray (N,3)
        Cartesian coordinates of points to act upon

    Returns
    -------
    ndarray (N,3)
        Cartesian coordinates of points after rotor application

    """
    return np.array(
        mp.Pool().map(
            partial(point_p_act_helper, points=points, param=param, poly=poly),
            range(points.shape[0]),
        )
    )


def point_cube_gen(
    center=np.array([0, 0, 0]), length=np.array([1, 1, 1]), subd=np.array([10, 10, 10])
):
    """Generate cube wtih

    Parameters
    ----------
    length : nparray, optional
        Edge lengths of cube.
    center : nparray, optional
        Center of cube.
    subd : nparray, optional
        subdivisions along axes.

    Returns
    -------
    points: nparray(N,3)
        Array of points generating cube.
    colors: nparray(N,4)
        Array of colors for plotting.

    """
    subd = np.array(subd, dtype=int)
    center = np.array(center)
    length = np.array(length)
    len_div = np.array(length)
    for i in range(3):
        if subd[i] == 1:
            length[i] = 0
        if length[i] == 0:
            len_div[i] = 1
    pointsx = np.linspace(0, length[0], subd[0])
    pointsy = np.linspace(0, length[1], subd[1])
    pointsz = np.linspace(0, length[2], subd[2])

    return point_cube_gen_help(pointsx, pointsy, pointsz, center, length, len_div, subd)


@njit(parallel=True, cache=True)
def point_cube_gen_help(pointsx, pointsy, pointsz, center, length, len_div, subd):
    """TODO: Docstring for point_cube_gen_help.
    Returns
    -------
    TODO

    """
    sx = subd[0]
    sy = subd[1]
    sz = subd[2]
    # set length to 0 if there is only one subdivision to assure correct center
    plot_points = np.zeros((sx * sy * sz, 3))
    colors = np.ones((sx * sy * sz, 4))
    center_off = np.array(
        [
            center[0] - length[0] / 2,
            center[1] - length[1] / 2,
            center[2] - length[2] / 2,
        ]
    )
    # set length to 1 if it is zero to avoid division by zero
    for x in range(sx):
        for y in range(sy):
            for z in range(sz):
                plot_points[sz * sy * x + sz * y + z, 0] = pointsx[x]
                plot_points[sz * sy * x + sz * y + z, 1] = pointsy[y]
                plot_points[sz * sy * x + sz * y + z, 2] = pointsz[z]
                colors[sz * sy * x + sz * y + z, 0] = 1 - (pointsx[x] / len_div[0])
                colors[sz * sy * x + sz * y + z, 1] = 1 - (pointsy[y] / len_div[1])
                colors[sz * sy * x + sz * y + z, 2] = 1 - (pointsz[z] / len_div[2])
    return (plot_points + center_off), colors


def sphere_gen(
    center=np.array([0, 0, 0]),
    radius=1,
    rows=10,
    cols=20,
    color=(0.5, 0.5, 0.5, 0.5),
    smooth=True,
):  # pragma: no cover
    """TODO: Docstring for sphere_gen.

    Parameters
    ----------
    center : ndarray (3), optional
        Center of sphere
    radius : float, optional
        Radius of sphere
    rows : int, optional
        Number of rows in sphere
    cols : int, optional
        Number of columns in sphere
    color: tuple (4), optional
        Default face colors
    smooth : boolean, optional
        Sould be shaded smooth

    Returns
    -------
    GLMeshItem
        Mesh of sphere


    """
    md = gl.MeshData.sphere(rows=rows, cols=cols)
    colors = np.ones((md.faceCount(), 4), dtype=float)
    md.setFaceColors(colors)
    m3 = gl.GLMeshItem(meshdata=md, smooth=smooth, shader="balloon")
    m3.translate(center[0], center[1], center[2])
    m3.scale(radius, radius, radius)
    m3.setColor(color)
    return m3
